/**********************************************************************
 * LeechCraft - modular cross-platform feature rich internet client.
 * Copyright (C) 2006-2014  Georg Rudoy
 *
 * Boost Software License - Version 1.0 - August 17th, 2003
 *
 * Permission is hereby granted, free of charge, to any person or organization
 * obtaining a copy of the software and accompanying documentation covered by
 * this license (the "Software") to use, reproduce, display, distribute,
 * execute, and transmit the Software, and to prepare derivative works of the
 * Software, and to permit third-parties to whom the Software is furnished to
 * do so, all subject to the following:
 *
 * The copyright notices in the Software and this entire statement, including
 * the above license grant, this restriction and the following disclaimer,
 * must be included in all copies of the Software, in whole or in part, and
 * all derivative works of the Software, unless such copies or derivative
 * works are solely in the form of machine-executable object code generated by
 * a source language processor.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE, TITLE AND NON-INFRINGEMENT. IN NO EVENT
 * SHALL THE COPYRIGHT HOLDERS OR ANYONE DISTRIBUTING THE SOFTWARE BE LIABLE
 * FOR ANY DAMAGES OR OTHER LIABILITY, WHETHER IN CONTRACT, TORT OR OTHERWISE,
 * ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
 * DEALINGS IN THE SOFTWARE.
 **********************************************************************/

#include "yandexdisk.h"
#include <algorithm>
#include <QIcon>
#include <QSettings>
#include <QCoreApplication>
#include <util/util.h>
#include "account.h"

namespace LeechCraft
{
namespace NetStoreManager
{
namespace YandexDisk
{
	void Plugin::Init (ICoreProxy_ptr proxy)
	{
		Util::InstallTranslator ("netstoremanager_yandexdisk");
	}

	void Plugin::SecondInit ()
	{
		ReadAccounts ();
	}

	void Plugin::Release ()
	{
	}

	QByteArray Plugin::GetUniqueID () const
	{
		return "org.LeechCraft.NetStoreManager.YandexDisk";
	}

	QString Plugin::GetName () const
	{
		return "NetStoreManager: Yandex.Disk";
	}

	QString Plugin::GetInfo () const
	{
		return tr ("Support for the Yandex.Disk for NetStoreManager plugin.");
	}

	QIcon Plugin::GetIcon () const
	{
		static QIcon icon ("lcicons:/netstoremanager/yandexdisk/resources/images/netstoremanager_yandexdisk.svg");
		return icon;
	}

	QSet<QByteArray> Plugin::GetPluginClasses () const
	{
		QSet<QByteArray> classes;
		classes << "org.LeechCraft.Plugins.NetStoreManager.Plugins.IStoragePlugin";
		return classes;
	}

	QObject* Plugin::GetQObject ()
	{
		return this;
	}

	QString Plugin::GetStorageName () const
	{
		return tr ("Yandex.Disk");
	}

	QIcon Plugin::GetStorageIcon () const
	{
		return QIcon (":/netstoremanager/yandexdisk/resources/images/yandexnarodlogo.png");
	}

	void Plugin::RegisterAccount (const QString& accName)
	{
		Account_ptr acc (new Account (this));
		acc->SetAccountName (accName);

		if (!acc->ExecConfigDialog ())
			return;

		Accounts_ << acc;
		WriteAccounts ();
		emit accountAdded (acc.get ());
	}

	QObjectList Plugin::GetAccounts () const
	{
		QObjectList result;
		Q_FOREACH (Account_ptr acc, Accounts_)
			result << acc.get ();
		return result;
	}

	void Plugin::RemoveAccount (QObject *obj)
	{
		auto pos = std::find_if (Accounts_.begin (), Accounts_.end (),
				[obj] (decltype (Accounts_.front ()) acc)
					{ return acc.get () == obj; });
		if (pos == Accounts_.end ())
			return;

		Accounts_.erase (pos);
		WriteAccounts ();
		emit accountRemoved (obj);
	}

	void Plugin::ReadAccounts ()
	{
		QSettings settings (QSettings::IniFormat, QSettings::UserScope,
				QCoreApplication::organizationName (),
				QCoreApplication::applicationName () + "_NSM_YD_Accounts");
		int size = settings.beginReadArray ("Accounts");
		for (int i = 0; i < size; ++i)
		{
				settings.setArrayIndex (i);
				const QByteArray& data = settings.value ("SerializedData").toByteArray ();
				Account_ptr acc = Account::Deserialize (data, this);
				Accounts_ << acc;
				emit accountAdded (acc.get ());
		}
		settings.endArray ();
	}

	void Plugin::WriteAccounts () const
	{
		QSettings settings (QSettings::IniFormat, QSettings::UserScope,
				QCoreApplication::organizationName (),
				QCoreApplication::applicationName () + "_NSM_YD_Accounts");
		settings.beginWriteArray ("Accounts");
		for (int i = 0; i < Accounts_.size (); ++i)
		{
			settings.setArrayIndex (i);
			settings.setValue ("SerializedData", Accounts_.at (i)->Serialize ());
		}
		settings.endArray ();
	}

	void Plugin::initPlugin (QObject *proxy)
	{
	}
}
}
}

LC_EXPORT_PLUGIN (leechcraft_netstoremanager_yandexdisk, LeechCraft::NetStoreManager::YandexDisk::Plugin);
